# -*- coding: utf8 -*-
'''
Created on 10.12.2010

@author: gsu
'''

from qpid.messaging import Connection, Message
import qpid
from threading import Thread
import pkgutil, inspect
import logging, logging.config
import time
from optparse import OptionParser
from contextlib import closing
from plugins_base import BasePlugin, ControlException, StopException, RestartException
import Queue
from daemon import createDaemon

logging.config.fileConfig('log.ini')
logger = logging.getLogger('su.artix.aqspy')

class PluginManager(object):
    
    plugins = []
    jobs = []
    
    def __init__(self, options, connection):
        self.control = Queue.Queue()
        self.options = options
        self.plugins = []
        self.connection = connection
        for importer, modname, _ in pkgutil.iter_modules(["plugins"]):
            try:
                pluginname = importer.path + '.' + modname
                logger.info(u'Подключаем плагин %s' % pluginname)
                self.plugins.append(__import__(pluginname, fromlist='q'))
            except Exception, e:
                logger.error(u'Ошибка подключения плагина %s (%s)' % (modname, e))
                logger.exception(e)

    def _stopAllJobs(self):
        for job in self.jobs:
            if job.isAlive():
                logging.info(u'Останавливаем работу %s' % job)
                job.exit()
                logging.info(u'Работа %s остановлена' % job)
        self.jobs = []

    def startAllJobs(self):
        self.jobs = []
        for plugin in self.plugins:
            for thing in map(lambda x: getattr(plugin, x), dir(plugin)):
                if inspect.isclass(thing) and issubclass(thing, BasePlugin) and thing != BasePlugin:
                    logger.info(u'Создаём работу для %s' % thing.__name__)
                    self.jobs.append(QueuePluginJob(thing(), self.connection, self.control, respawn=True, respawnTimeout=10.0))
        for job in self.jobs:
            logger.info(u'Запуск работы %s' % job)
            job.start()

    def manageJobs(self):
        while True:
            action = self.control.get(block=True)
            if action == 'stop':
                self._stopAllJobs()
                break
            if action == 'restart':
                self._stopAllJobs()
                self.startAllJobs()

class QueuePluginJob(Thread):

    plugin = None

    def __init__(self, plugin, connection, control, respawn=False, respawnTimeout=10.0):
        logger.info(u'Инициализация потока обработки очереди')
        super(QueuePluginJob, self).__init__()
        self.control = control
        self.plugin = plugin
        self.connection = connection        
        self.respawn = plugin.respawn if hasattr(plugin, "respawn") else respawn 
        self.respawnTimeout = plugin.respawnTimeout if hasattr(plugin, "respawnTimeout") else respawnTimeout      

    def keepPlugin(self, connection, plugin):
        while True:
            logger.info(u'Запуск плагина %s' % plugin)
            try:
                self.plugin.run(connection)
                break
            except ControlException, e:
                logger.info(u'Управляющее исключение %s' % e)
                raise e
            except Exception, e:
                logger.info(u'Плагин упал с ошибкой %s:%s' % (type(e), e))
                logger.exception(e)
                if not self.respawn:
                    break
                else:
                    logger.info(u'Перезапуск через %f секунд' % self.respawnTimeout)
                    time.sleep(self.respawnTimeout)

    def exit(self):
        self.plugin.exit = True
        self.join()

    def run(self):
        try:
            self.keepPlugin(self.connection, self.plugin)
        except StopException:
            self.control.put('stop', block=True)
        except RestartException:
            self.control.put('restart', block=True)

    def __str__(self):
        return u'%s' % str(self.plugin)

class Main():

    stdin = '/dev/null'
    stdout = '/dev/null'
    stderr = '/dev/null'
    
    def __init__(self):

        self.parser = OptionParser("usage: %prog [options] start|stop|restart")
        self.parser.add_option('-b', '--broker',
                          metavar='BROKER_URL', default='localhost:5672',
                          help='set broker url')
        self.parser.add_option('-d', '--daemon', action='store_true', default=False,
                        help='start as daemon')

        self.options, self.args = self.parser.parse_args()
        if len(self.args) != 1:
            self.parser.print_help()

    def fetchAll(self, session, receiver):
        reply = True
        while reply:
            try:
                reply = receiver.fetch(1)
                session.acknowledge()
            except Exception, e:
                logger.info(u'error: %s/%s', type(e), e)
                reply = None
    
    def sendControlMessage(self, msg, connection):
        session = connection.session()
        control_addr = 'artix.aqspy.control; {create:always}'
        resp_addr = "artix.aqspy.control_response; {create:always}"
        sender = session.sender(control_addr)
        receiver = session.receiver(resp_addr)
        control_receiver = session.receiver(control_addr)
        self.fetchAll(session, control_receiver)
        self.fetchAll(session, receiver)
        control_receiver.close()
        request = Message(msg, reply_to=resp_addr)
        sender.send(request)
        reply = None
        try:
            reply = receiver.fetch(2.0)
            logger.info(u'Получен Ответ на управляюще сообщение: %s' % reply.content)
            self.fetchAll(session, receiver)
        except:
            pass
        finally:
            session.acknowledge()
#            session.close()
        return reply.content if reply else None
    
    def testAlive(self, connection):
        try:
            return self.sendControlMessage('ping', connection)
        except Exception, e:
            logger.error(u'Не получилось отправить управляющее сообщение ping')
            logger.error(u'Ошибка: %s' % e)
            return False
    
    def start(self):
        connection = Connection(self.options.broker, reconnect=True, reconnect_interval=10.0)
        connection.open()
        if self.testAlive(connection):
            logging.info(u'Обработчик очереди уже запущен!')
            return
        connection.close()
        connection.open()
        pluginManager = PluginManager(self.options, connection)
        pluginManager.startAllJobs()
        pluginManager.manageJobs()

    def stop(self):
        connection = Connection(self.options.broker)
        with closing(connection):
            connection.open()
            self.sendControlMessage('stop', connection)

    def restart(self):
        connection = Connection(self.options.broker)
        with closing(connection):
            connection.open()
            self.sendControlMessage('restart', connection)

    def main(self):
        if 'start' in self.args:
            logger.info(u'Запуск обработчика очередей')
            if self.options.daemon:
                logging.shutdown()
                createDaemon()
                logging.config.fileConfig('log.ini')
            self.start()
        if 'stop' in self.args:
            logger.info(u'Останов обработчика очередей')
            self.stop()
        if 'restart' in self.args:
            logger.info(u'Перезапуск обработчика очередей')
            self.restart()

if __name__ == '__main__':
    Main().main()
